// pecco -- please enjoy classification with conjunctive features
//  $Id: classify.cpp 474 2010-08-13 16:17:00Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include "classify.h"

#define SHIFT_LEFT_UNSIGNED_USES_SHL (((unsigned int)0xffffffff >> 1) == 0x7fffffff)

namespace pecco {
  
  template <feat_t N>
  bool ClassifierBase <N>:: _setOpt (const char* opt_str) {
    if (_verbosity > 0) std::cerr << opt_str << std::endl;
    char* p = new char[std::strlen (opt_str) + 1];
    std::strcpy (p, opt_str);
    std::vector <std::string> argvec;
    char* q = p;
    for (p = strtok(p, " "); p != NULL; p = strtok(NULL, " "))
      argvec.push_back (std::string (p));
    delete [] q;
    _opt.parse (argvec);
    
    _verbosity = ny::strton <int> (_opt["-v"].c_str ());
    if      (_opt["-zt"] == "none")  _zmode = NONE;
    else if (_opt["-zt"] == "nf")    _zmode = NFREQ;
    else if (_opt["-zt"] == "nff")   _zmode = NFREQ_FCNT;
    else if (_opt["-zt"] == "naive") _zmode = NAIVE;
    else ny::print_err (HERE "unknown compression method.\n");
    if (_verbosity > 0) _opt.print ();
    return true;
  }
  // bytewise coding (cf. Williams & Zobel, 1999)
  template <feat_t N>
  inline void ClassifierBase <N>::_vbEncode (ny::uint fi, ny::uchar* s, ny::uint &len) {
    for (s[len] = (fi & 0x7f); fi >>= 7; s[++len] =  (fi & 0x7f))
      s[len] |= 0x80;
    ++len;
  }
  // radix sort for very long keys
  template <feat_t N>
  template <typename value_type, typename SrcIterator, typename OutIterator>
  inline void ClassifierBase <N>::_radix (SrcIterator first, SrcIterator last,
                                          OutIterator dest, size_t shift) {
    value_type count[BIN_ELM];
    std::memset (count, 0, sizeof (count));
    for (SrcIterator it = first; it != last; ++it)
      ++count[((*it)>>shift) & BIN_MAX];
    value_type offset[BIN_ELM];
    offset[0] = 0;
    std::partial_sum (&count[0], &count[(BIN_ELM)-1], &offset[1]);
  
    for (SrcIterator it = first; it != last; ++it)
      *(dest + offset[((*it)>>shift) & BIN_MAX]++) = *it;
  }

  template <feat_t N>
  template <size_t SIZE, typename Iterator>
  inline void ClassifierBase <N>::_radix_sort (const Iterator &first, const Iterator &last)
  {
    typedef typename std::iterator_traits<Iterator>::value_type value_type;
    if (first == last) return;
    size_t n = last - first;
    value_type temp[n];
    for (size_t i = 0; i < SIZE / BIN; ++i) {
      _radix <value_type> (first, last, &temp[0], i*BIN);
      _radix <value_type> (&temp[0], &temp[n], first, ++i*BIN);
    }
  }
  // for short feature vector < 50
  template <feat_t N>
  template <typename Iterator>
  inline void ClassifierBase <N>::_insertion_sort (const Iterator &first,
                                                   const Iterator &last) {
    typedef typename std::iterator_traits <Iterator>::value_type value_type;
    std::less <value_type> cmp;
    if (first == last) return;
    Iterator sorted = first;
    for (++sorted; sorted != last; ++sorted) {
      value_type temp = *sorted;
      Iterator prev, curr;
      prev = curr = sorted;
      for (--prev; curr != first && cmp (temp, *prev); --prev, --curr)
        *curr = *prev;
      *curr = temp;
    }
  }
  // Zipf-aware sorting funciton
  template <feat_t N>
  template <typename T, typename Iterator>
  inline void ClassifierBase <N>::_bucket_sort (const Iterator &first, const Iterator &last)
  {
    typedef typename std::iterator_traits <Iterator>::value_type value_type;
  
    if (first == last) return;
    T bucket = 0;
    // assuming most feature IDs are less than sizeof (T) * 8
    Iterator it, jt;
    for (it = jt = last; it > first;) {
      const value_type &fi = *--it;
      if (fi < sizeof (T) * 8)
        bucket |= (static_cast <T> (1) << fi);
      else // put unsorted (large) feature IDs on tail
        *--jt = fi;
    }
    // do bucket sort
    it = first;
    value_type fi = 1;
    while (bucket >>= 1)
      { if (bucket & static_cast <T> (1)) { *it = fi; ++it; }; ++fi; }
    // do insertion sort for the tail
    _insertion_sort (it, last);
  }
  template <feat_t N>
  void ClassifierBase <N>::_convertCfstr2Cf (char* cs, ny::fv_t &fv, const bool flag) {
    // convert a feature number string to feature indices
    if (! flag) fv.reserve (_maf);
    for (cs = std::strtok (cs, ":"); cs != NULL; cs = std::strtok (NULL, ":")) {
      ny::uint fn = ny::strton <ny::uint> (cs);
      if (flag) {
        if (_fncnt.find (fn) == _fncnt.end ())
          _fncnt.insert (counter_t::value_type (fn, 0));
        fv.push_back (fn);
      } else {
        if (fn < _fn2fi.size () && _fn2fi[fn] != 0) fv.push_back (_fn2fi[fn]);
      }
    }
    if (! flag) std::sort (fv.begin (), fv.end ());
  }
  template <feat_t N>
  void ClassifierBase <N>::_convertEx2Fv (char* cs, ny::fv_t &fv, const bool flag) {
    // convert an example string in SVM-light format into feature indices
    if (! flag) fv.reserve (_maf);
    for (cs = std::strtok (cs, ": \t"); cs != NULL; cs = std::strtok (NULL, ": \t")) {
      ny::uint fn = ny::strton <ny::uint> (cs);
      if (flag) {
        if (_fncnt.find (fn) == _fncnt.end ()) 
          _fncnt.insert (counter_t::value_type (fn, 0));
        fv.push_back (fn);
      } else {
        if (fn < _fn2fi.size () && _fn2fi[fn] != 0) fv.push_back (_fn2fi[fn]);
      }
      cs = std::strtok (NULL, ": \t");
    }
    if (! flag) std::sort (fv.begin (), fv.end ());
  }
  template <feat_t N>
  void ClassifierBase <N>::_convertFv2Fv (ny::fv_t &fnv) const {
    // convert feature numbers to feature indices
    const ny::fv_t::iterator it_end = fnv.end ();
    ny::fv_t::iterator jt = fnv.begin ();
    for (ny::fv_it it = jt; it != it_end; ++it) {
      const ny::uint fn = *it;
      if (fn < _fn2fi.size ())
        if (ny::uint fi = _fn2fi[fn]) { *jt = fi; ++jt; }
    }
    fnv.erase (jt, it_end);
  }
  //
  template <feat_t N>
  void ClassifierBase <N>::_sortFv (ny::fv_t &fnv) const {
#ifdef SHIFT_LEFT_UNSIGNED_USES_SHL
    // for feature vectors following Zipf
    _bucket_sort <unsigned int> (fnv.begin (), fnv.end ());
#else
    // for general feature vectors
    std::sort (fnv.begin (), fnv.end ());
    // for general short feature vectors
    // _insertion_sort (fnv.begin (), fnv.end ());
    // for general long feacture vectors
    // _radix_sort <SORT_MAX> (fnv.begin (), fnv.end ());
#endif
  }
  // * a function that estimates the additional cost
  //   when eliminating nodes
  template <feat_t N>
  inline size_t ClassifierBase <N>::_cost_fun (size_t m, size_t n) {
    // by frequency * # of features
    switch (_d) {
      case 1: return 0;
      case 2: return ((n*n - n) - (m*m - m)) / 2;
      case 3: return ((n*n*n - n) - (m*m*m - m)) / 6;
      case 4: return ((n*n*n*n - 2*n*n*n + 11*n*n - 10*n)
                      -(m*m*m*m - 2*m*m*m + 11*m*m - 10*m)) / 24;
      default: assert (! "please add case statement."); return 0; // dummy;
    }
  }
  template <feat_t N>
  inline const ny::fv_it ClassifierBase <N>::_splitFv (const ny::fv_t &fv) const
  { return std::lower_bound (fv.begin (), fv.end (), _f_r); }
  // assign younger feature indices to important feature numbers
  template <feat_t N>
  bool ClassifierBase <N>::_renameFeatureID (std::vector <ny::fv_t> &fvv) {
    if (_verbosity > 0) std::cerr << "Renaming feature id..";
#ifdef USE_FREQ_SORT
    // rename features according to frequency in training data
    const std::string tfn (_opt["-lf"]); // or cache origin data??
    FILE* reader =  std::fopen (tfn.c_str (), "r");
    if (! reader) ny::print_err (HERE "no such file: %s\n", tfn.c_str ());
    _nt         = 0;
    char*  line = 0;
    size_t read = 0;
    while (ny::getLine (reader, line, read)) {
      if (*line != '\0') {
        *(line + read - 1) = '\0';
        // char* p_end = line + read - 1;
        // assume primitive feature vectors
        std::strtok (line, " :\t");
        for (char* p = std::strtok (NULL, " :\t");
             p != NULL; p = std::strtok (NULL, " :\t")) {
          std::strtok (NULL, " :\t");
          // for (char* p = ny::skipChar (line, ' ');
          //      p <= p_end; p = ny::skipChar (p, ' ', p_end)) {
          // char* q = ny::skipChar (p, ':');
          const ny::uint fn = ny::strton <ny::uint> (p);
          counter_t::iterator it = _fncnt.find (fn);
          if (it != _fncnt.end ()) it->second += 1;
        }
        ++_nt;
      }
    }
    std::fclose (reader);
#else
    // rename features according to # support vectors / conjunctive features
    for (std::vector <ny::fv_t>::const_iterator it = fvv.begin ();
         it != fvv.end (); ++it) // for each unit
      for (ny::fv_it fit = fit->begin (); fit != fit->end (); ++fit) {
        counter_t::iterator jt = _fncnt.find (*fit); // for each feature
        if (jt != _fncnt.end ()) jt->second += 1;
      }
#endif
    // building feature mapping
    ny::counter <ny::uint>::type fn_counter;
    fn_counter.reserve (_fncnt.size ());
    ny::uint fn_max = 0;
    for (counter_t::const_iterator it = _fncnt.begin ();
         it != _fncnt.end (); ++it) {
      const ny::uint fn  = it->first;
      const ny::uint cnt = it->second;
      fn_max = std::max (fn, fn_max);
      fn_counter.push_back (ny::counter <ny::uint>::type::value_type (cnt, fn));
    }
    std::sort (fn_counter.begin (), fn_counter.end ());
    _fn2fi.resize (fn_max + 1, 0);
    _fi2fn.resize (_nf + 1, 0); // _nf
    ny::uint fi = 1;
    for (ny::counter <ny::uint>::type::reverse_iterator it = fn_counter.rbegin ();
         it != fn_counter.rend (); ++it) {
      const ny::uint fn = it->second;
      _fi2fn[fi] = fn;
      _fn2fi[fn] = fi;
      ++fi;
    }
    // rename features in conjunctive features / support vectors
    for (std::vector <ny::fv_t>::iterator it = fvv.begin ();
         it != fvv.end (); ++it) { // for each unit
      _convertFv2Fv (*it);
      _sortFv (*it); // for pkb; bug
    }
    if (_verbosity > 0) std::cerr << "done." << std::endl;
    return true;
  }
  
  template <feat_t N>
  inline void ClassifierBase <N>::_addScore (ny::fl_t* score, int n)
  { for (ny::uint i = 0; i < _nl; ++i) score[i] += _feat_weight[n+i]; }
  template <>
  inline void ClassifierBase <BINARY>::_addScore (ny::fl_t* score, int n)
  { union byte_4 b4 (n); b4.u <<= 1; *score += b4.f; }

  template <feat_t N>
  inline int ClassifierBase <N>::_lookupFeat (const ny::fv_it &it, size_t &pos) {
    ny::uchar key[KEY_SIZE];
    ny::uint len = 0;
    _vbEncode (*it, key, len);
    size_t p = 0;
    int n = _pke.traverse (reinterpret_cast <const char*> (key), pos, p, len);
#ifdef USE_PROFILE
    ++_traverse; if (n >= 0) ++_lookup;
#endif
    return n;
  }
  
  template <feat_t N>
  template <int D>
  inline void ClassifierBase <N>::_pkeInnerLoop (ny::fl_t* score, ny::fv_it it, const ny::fv_it beg, const ny::fv_it end, const size_t pos) {
    for (; it != end; ++it) {
      size_t pos_ = pos;
      int n = _lookupFeat (it, pos_);
      if (n >= 0) _addScore (score, n);
      else if (n == -2) continue;
      // if (n != -2)
      _pkeInnerLoop <D-1> (score, beg, beg, it, pos_);
    }
  }
  // explicit specializations
  // template <>
  // template <>
  // inline void ClassifierBase <BINARY>::_pkeInnerLoop<0> (ny::fl_t*, ny::fv_it, const ny::fv_it, const ny::fv_it, const size_t) {}
  // template <>
  // template <>
  // inline void ClassifierBase <MULTI>::_pkeInnerLoop<0>  (ny::fl_t*, ny::fv_it, const ny::fv_it, const ny::fv_it, const size_t) {}
  template <>
  template <>
  inline void ClassifierBase <BINARY>::_pkeInnerLoop <1> (ny::fl_t* score, ny::fv_it it, const ny::fv_it, const ny::fv_it end, const size_t pos) {
    for (; it != end; ++it) {
      size_t pos_ = pos;
      int n = _lookupFeat (it, pos_);
      if (n >= 0) _addScore (score, n);
    }
  }
  template <>
  template <>
  inline void ClassifierBase <MULTI>::_pkeInnerLoop <1> (ny::fl_t* score, ny::fv_it it, const ny::fv_it, const ny::fv_it end, const size_t pos) {
    for (; it != end; ++it) {
      size_t pos_ = pos;
      int n = _lookupFeat (it, pos_);
      if (n >= 0) _addScore (score, n);
    }
  }

  template <feat_t N>
  ny::fl_t* ClassifierBase <N>::_pkeClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it &end) {
    const ny::fv_it beg = fv.begin ();
#ifdef USE_PROFILE
    _flen += end - beg;
#endif
    switch (_d) {
      case 1: _pkeInnerLoop <1> (score, it, beg, end, 0); break;
      case 2: _pkeInnerLoop <2> (score, it, beg, end, 0); break;
      case 3: _pkeInnerLoop <3> (score, it, beg, end, 0); break;
      case 4: _pkeInnerLoop <4> (score, it, beg, end, 0); break;
      default: assert (! "please add case statement.");
    }
    return score;
  }
  
  template <feat_t N>
  ny::fl_t* ClassifierBase <N>::_fstClassify (const ny::fv_t &fv, ny::fl_t* score) {
    ny::uchar key[KEY_SIZE];
    int n = -1;
    ny::fv_it cit = fv.begin ();
    size_t   pos  = 0;
    ny::uint prev = 0;
    for (; cit < fv.end (); ++cit) {
      ny::uint len = 0;
      _vbEncode (*cit - prev, key, len);
      prev = *cit;
      size_t p = 0;
#ifdef USE_PROFILE
      ++_traverse_sp;
#endif
      int npp = _fst.traverse (reinterpret_cast <const char*> (key), pos, p, len);
      if (npp >= 0) n = npp;
      else break;
    }
#ifdef USE_PROFILE
    _all += _cost_fun (0, fv.end () - fv.begin ());
    _hit += _cost_fun (cit - fv.begin (),  fv.end () - fv.begin ());
    ++_lookup_sp;
#endif
    if (n >= 0) {
      if (N == BINARY) {
        union byte_4 b4 (n); b4.u <<= 1; *score += b4.f;
      } else {
        for (ny::uint i = 0; i < _nl; ++i) score[i] += _fst_weight[n+i];
      }
      if (cit == fv.end ()) return score;
    }
    // return _pkeClassify (fv, score, cit, end); // used in EMNLP
    return _baseClassify (fv, score, cit, fv.end ()); // splitSVM + FST
  }
  //
  template <feat_t N>
  std::string ClassifierBase <N>::_setFStrieFN (const std::string &fn_base, const std::string &zfact) {
    std::string fn (fn_base);
    switch (_zmode) {
      case NFREQ:      fn += "_nf.-"  + zfact; break;
      case NFREQ_FCNT: fn += "_nff.-" + zfact; break;
      case NAIVE:      fn += "_niv.-" + zfact; break;
      case NONE: break;
    }
    return fn;
  }
  // fstrie construction
  template <feat_t N>
  bool ClassifierBase <N>::_setFStrie () {
    const bool abuse = _abuse_double_array ();
    const std::string cfn   (_opt["-cf"]);
    const std::string zfact (_opt["-z"]);
    std::string dafn_base (cfn + (abuse ? "." : ".n") + ny::DA_SUFFIX);
    std::string alfn_base (cfn + ".alpha");
    std::string dafn (_setFStrieFN (dafn_base, zfact));
    std::string alfn (_setFStrieFN (alfn_base, zfact));
    //  check the update time of events
    if (_verbosity > 0) std::cerr << "Loading fstrie..";
    if (newer (dafn.c_str (), cfn.c_str ()) && _fst.open (dafn.c_str ()) == 0) {
      if (! abuse) { // load the pre-computed weights
        FILE* reader = std::fopen (alfn.c_str (), "rb");
        if (! reader) ny::print_err (HERE "no such file: %s\n", alfn.c_str ());
        if (std::fseek (reader, 0, SEEK_END) != 0) return -1;
        ny::uint nfs = std::ftell (reader) / (_nl * sizeof (ny::fl_t));
        if (std::fseek (reader, 0, SEEK_SET) != 0) return -1;
        _fst_weight = new ny::fl_t [_nl*nfs];
        std::fread (&_fst_weight[0], sizeof (ny::fl_t), _nl * nfs, reader);
        std::fclose (reader);
      }
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    } else {
      if (_verbosity > 0) std::cerr << "not found." << std::endl;
      FILE* reader = std::fopen (cfn.c_str (), "r");
      if (! reader) ny::print_err (HERE "no such event file: %s\n", cfn.c_str ());
      if (_verbosity > 0)
        std::cerr << "Building fstrie from events.." << std::endl;
      ny::uint nt   = 0;
      size_t   nkey = 0;
      std::set <FstKey*, FeatKeypLess> fst_key;
      std::vector <ny::uchar> s;
      ny::fl_t * w = new ny::fl_t[_nl];
      char*  line = 0;
      size_t read = 0;
      while (ny::getLine (reader, line, read)) {
        if (*line != '\0') { // skip label
          // char* p = ny::skipChar (line, ' ');
          char* p = line; while (*p != '\0' && *p != '\t' && *p != ' ') ++p; ++p;
          _fv.clear ();
          _convertEx2Fv (p, _fv, false);
          const size_t fsize = _fv.size ();
          // const size_t fsize =
          //   std::min (_fv.size (),
          //             static_cast <size_t> (_splitFv (_fv) - _fv.begin ()));
          // precompute score
          ny::uint l    = 0;
          ny::uint prev = 0;
          std::fill (&w[0], &w[_nl], 0);
          s.reserve (KEY_SIZE * fsize);
          for (size_t j = 1; j <= fsize; ++j) {
            ny::uint n = 0;
            _vbEncode (_fv[j-1] - prev, &s[0] + l, n); // use diff to save space
            prev = _fv[j-1];
            l += n;
            FstKey* q = new FstKey (&s[0], 0, l, 0);
            std::set <FstKey*, FeatKeypLess>::iterator it
              = fst_key.lower_bound (q);
            if (it == fst_key.end () || FeatKeypLess () (q, *it)) {
              ++nkey;
              _baseClassify (_fv, &w[0], _fv.begin () + j - 1, _fv.begin() + j);
              it = fst_key.insert (it, q);
              std::memcpy ((*it)->cont, &w[0], _nl * sizeof (ny::fl_t));
              switch (_zmode) {
                case NFREQ:      (*it)->weight = 1;                     break;
                case NFREQ_FCNT: (*it)->weight = _cost_fun (j-1, j);    break;
                case NAIVE:      (*it)->weight = nt; (*it)->count = 1; break;
                case NONE: break;
              }
            } else { // store the impact of the nodes on the speed-up
              // assert ((*it)->tail == false || j == fsize);
              std::memcpy (&w[0], (*it)->cont, _nl * sizeof (ny::fl_t));
              delete q;
            }
            if (j == fsize)      (*it)->tail = true; // maybe incorrect
            if (_zmode != NAIVE) (*it)->count += 1;
          }
        }
        if ( ++nt % 1000 == 0 && _verbosity > 0) {
          std::cerr << "\r processing event.. " << nt << " => ";
          std::cerr << nkey << " keys generated";
        }
      }
      delete [] w;
      std::fclose (reader);
      if (_verbosity > 0) {
        std::cerr << "\r processing event..done. " << nt << " event => ";
        std::cerr << nkey << " keys generated" << std::endl;
      }
      std::set <FstKey*, FstKeypLess> fst_key_sort;
      for (std::set <FstKey*, FstKeypLess>::const_iterator it = fst_key.begin ();
           it != fst_key.end (); ++it)
        if (_zmode == NAIVE || (*it)->tail) fst_key_sort.insert (*it);
      std::vector <const char*> str;
      std::vector <size_t>      len;
      std::vector <int>         val;
      str.reserve (nkey);
      len.reserve (nkey);
      val.reserve (nkey);
      if (! abuse)
        _fst_weight = new ny::fl_t [_nl*nkey];
      size_t nkey_small = nkey;
      size_t nt_small   = nt;
      for (size_t j = 0; (_zmode != NAIVE || nt_small > 0) && nkey_small > 0; ++j) {
        // sort dictionary order of key strings
        if (_zmode == NAIVE) { // use initial N
          std::set <FstKey*, FstKeypLess>::const_iterator it = fst_key_sort.begin ();
          while (it != fst_key_sort.end () && (*it)->weight < nt_small) ++it;
          while (it != fst_key_sort.end ()) {
            FstKey *p = *it; ++it;
            if (it != fst_key_sort.end ()) {
              FstKey *q = *it;
              fst_key_sort.erase (p);
              it = fst_key_sort.find (q);
            } else {
              fst_key_sort.erase (p);
            }
            fst_key.erase (p);
            delete p;
          }
        } else {
          while (fst_key.size () > nkey_small) {
            // should memory release
            std::set <FstKey*, FstKeypLess>::iterator it = fst_key_sort.begin ();
            FstKey* p = *it;
            std::set <FstKey*, FeatKeypLess>::iterator dit = fst_key.find (p);
            // warning: a leaf node is sorted by frequency and look left/right
            // 1 2 3
            // 1 2 3 4 - x1
            // 1 2 3 5
            // 1 2 3 5 6 -
            if (dit != fst_key.begin ()) { // not root node
              std::set <FstKey*, FeatKeypLess>::iterator dit_ = dit;
              FstKey* r = *(--dit_); ++dit_;
              if (r->is_prefix (p)) // prefix
                if (++dit_ == fst_key.end () || ! r->is_prefix (*dit_))
                  fst_key_sort.insert (it, r); // no sibling
            }
            fst_key_sort.erase (it);
            fst_key.erase (dit);
            delete p;
          }
        }
        ny::uint i = 0;
        std::set <FstKey*, FeatKeypLess>::const_iterator dit = fst_key.begin ();
        for (; dit != fst_key.end (); ++dit) {
          FstKey* p = *dit;
          str.push_back (reinterpret_cast <char*> (p->key));
          len.push_back (p->len);
          ny::fl_t* wv = p->cont;
          if (abuse) {
            union byte_4 b4 (static_cast <float> (*wv));
            b4.u >>= 1;
            val.push_back (b4.i);
          } else {
            val.push_back (i*_nl);
            for (size_t li = 0; li < _nl; ++li)
              _fst_weight[_nl * i + li] = wv[li]; // 0:(li=0, i=0) 1:(li=1, i=0)
          }
          ++i;
        }
        std::ostringstream ss;
        ss << j; // zfact
        if (!abuse) {
          std::string alfn_tmp (_setFStrieFN (alfn_base, ss.str ()));
          FILE* writer = std::fopen (alfn_tmp.c_str (), "wb");
          std::fwrite (&_fst_weight[0], sizeof (ny::fl_t), i *_nl, writer);
          std::fclose (writer);
        }
        std::string dafn_tmp (_setFStrieFN (dafn_base, ss.str ()));
        ss.str ("");
        ss << "fstrie (compression rate 2^-" << j << ")";
        buildDA (&_fst, ss.str (), dafn_tmp, str, len, val, _verbosity > 0);
        _fst.clear ();
        str.clear  ();
        len.clear  ();
        val.clear  ();
        nt_small   >>= 1;
        nkey_small >>= 1;
        if (_zmode == NONE) break;
      }
      // try reload
      if (_fst.open (dafn.c_str ()) != 0)
        ny::print_err (HERE "no such double array: %s\n", dafn.c_str ());
      if (!abuse) {
        delete [] _fst_weight;
        // load computed score
        FILE* reader = std::fopen (alfn.c_str (), "rb");
        if (std::fseek (reader, 0, SEEK_END) != 0) return -1;
        ny::uint nfs = std::ftell (reader) / (_nl * sizeof (ny::fl_t));
        if (std::fseek (reader, 0, SEEK_SET) != 0) return -1;
        _fst_weight = new ny::fl_t [_nl*nfs];
        std::fread (&_fst_weight[0], sizeof (ny::fl_t), _nl * nfs, reader);
        std::fclose (reader);
      }
      for (std::set <FstKey*>::const_iterator dit = fst_key.begin ();
           dit != fst_key.end (); ++dit)
        delete *dit;
      std::set <FstKey*, FeatKeypLess>().swap (fst_key);      // memory release
      std::set <FstKey*, FstKeypLess> ().swap (fst_key_sort); // memory release
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    }
    return true;
  }
#ifdef USE_PROFILE
  template <feat_t N>
  const void ClassifierBase <N>::printHitRatio () const {
    std::cerr << "DA flen: " << _flen << std::endl;
    if (_mode == SPLIT || _mode == FST || _mode == PKE) {
      std::cerr << "DA lookup: " << _lookup << std::endl;
      std::cerr << "DA traverse: " << _traverse << std::endl;
      if (_mode == FST) {
        std::cerr << "DA lookup (split): " << _lookup_sp << std::endl;
        std::cerr << "DA traverse (split): " << _traverse_sp << std::endl;
        std::cerr << "node reduction: " << _hit * 1.0 / _all
                  << " (" << _hit << "/" << _all << ")" << std::endl;
      }
    }
  }
#endif
  // explicit specialization
  template class ClassifierBase <BINARY>;
  template class ClassifierBase <MULTI>;
}
