// pecco -- please enjoy classification with conjunctive features
//  $Id: classify.h 472 2010-08-13 13:43:20Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef CLASSIFY_H
#define CLASSIFY_H

#include <sys/stat.h>
#include <cmath>
#include <cassert>
#include <vector>
#include <iostream>
#include <numeric>
#include <sstream>
#include <string>
#include <set>
#include "typedef.h"
#ifdef USE_HASH
#include <tr1/unordered_map>
#else
#include <map>
#endif
#ifdef USE_TIMER
#include "timer.h"
#endif
#include "opt_parse.h"

namespace pecco {
  // constant
  enum feat_t { BINARY, MULTI };
  // type alias
  // uchar* -> fl_t  ; conj. feat. -> weight (float)
  typedef ny::DAKeyBase  <ny::uchar, ny::fl_t> FeatKey;
  typedef ny::DAKeypLess <ny::uchar, ny::fl_t> FeatKeypLess;
  // uchar* -> fl_t  ; feat. seq.  -> weight (float) or weight ID (int)
  struct FstKey : public FeatKey {
    size_t   weight; // used for node cutoff
    ny::uint count;  // used for node cutoff
    bool     tail;
    FstKey (ny::uchar* k, ny::fl_t* c, size_t l, ny::uint nl = 1)
      : FeatKey (k, c, l, nl), weight (0), count (0), tail (false) {};
    bool is_prefix (FstKey *a) const { // whether key is a->key's prefix 
      if (len > a->len)
        return false;
      else
        for (size_t i = 0; i < len; ++i)
          if (key[i] != a->key[i]) return false;
      return true;
    }
  };
  // feature sequence ordering; see Yoshinaga and Kitsuregawa (EMNLP 2009)
  struct FstKeypLess {
    bool operator () (const FstKey* a, const FstKey* b) const {
      // keep frequent / long keys
      if (a->count * a->weight < b->count * b->weight)
        return true;
      else if (a->count * a->weight > b->count * b->weight)
        return false;
      // keep keys with frequent features
      else return FeatKeypLess () (b, a);
      // return (std::memcmp (a->key, b->key, a->len) > 0);
    }
  };
  // build / save da at once
  static inline void buildDA (ny::DA* da,
                              const std::string name,
                              const std::string &fn,
                              std::vector <const char*> &str,
                              const std::vector <size_t> &len,
                              const std::vector <ny::DA::result_type> &val, // int
                              bool flag, const char* mode = "wb") {
    if (flag) std::cerr << " Building " << name << " DA..";
    if (da->build (str.size (), &str[0], &len[0], &val[0]) != 0 ||
        da->save  (fn.c_str (), mode) != 0 ) {
      ny::print_err (HERE "failed to build %s DA.\n", name.c_str ());
    } else if (flag) {
      std::cerr << "done." << std::endl;
    }
  }
  // check file refreshness
  static inline bool newer (const char* newer, const char* base) {
    struct stat newer_fs, base_fs;
    if (stat (newer, &newer_fs) == -1) return false;
    stat (base,  &base_fs); // need not exist
    return difftime (newer_fs.st_mtime, base_fs.st_mtime) > 0;
  }
  template <feat_t N>
  class ClassifierBase : private ny::Uncopyable {
  protected:
    static const size_t KEY_SIZE = 5;         // >= log_{2^7} _nf + 1
    static const int    SORT_MAX = 32;        // 2^SORT_MAX must be > _nf
    static const int    BIN      = 6;         // BIN for radix sort
    static const int    BIN_ELM  = 1 << BIN;  // MAX for bucket sort
    static const int    BIN_MAX  = ((1 << BIN) - 1);
    // type alias
    typedef ny::map <ny::uint, ny::uint>::type counter_t;
    enum classifier_mode  {PKB, PKBPP, PKI, PKE, SPLIT, FST};
    enum compression_mode {NONE, NFREQ, NFREQ_FCNT, NAIVE};
    union byte_4 { // int <-> uint <-> float converter
      int i;
      ny::uint u;
      float f;
      byte_4 (int n)   : i (n) {};
      byte_4 (float b) : f (b) {};
    };
    // options and classifier parameters
    ny::fv_t         _fv;
    ny::Options      _opt;
    classifier_mode  _mode;
    int              _verbosity;
    ny::uint         _d;        // degree of feature combination
    ny::uint         _nl;       // # of labels
    compression_mode _zmode;    // fstrie compression type
    // various sizes
    ny::uint _nf;     // # active feature
    ny::uint _nf_cut; // # active feature (pruned)
    ny::uint _nt;     // # training sample
    ny::uint _nunit;  // # support vectors / conjunctive features
    // for mapping feature numbers to feature indices
    std::vector <ny::uint> _fn2fi;  // feature number -> feature index
    std::vector <ny::uint> _fi2fn;  // feature index  -> feature number
    counter_t              _fncnt;  // feature number -> count
    // double arrays for conjunctive features and fstrie
    ny::DA _pke;      // conjunctive feature -> weight / weight_id
    ny::DA _fst;      // feature sequece     -> weight / weight_id
    // temporary variables
    ny::fl_t*  _feat_weight; // conjunctive feature id -> weight
    ny::fl_t*  _fst_weight;  // feature sequence id    -> weight
    ny::uint   _f_r;  // min feature index of rare feature (= # common feature)
    ny::uint   _maf;         // # max active feature per vector;
    // timer
#ifdef USE_TIMER
    ny::TimerPool _timer_pool;
    ny::Timer*    _enc_t;      // feature mapping
    ny::Timer*    _model_t;    // compiling/loading  model
    ny::Timer*    _pke_t;      // pke classify
    ny::Timer*    _fst_t;      // fst classify
#endif
    // profiling
#ifdef USE_PROFILE
    ny::uint _all;         // all       (for split)
    ny::uint _hit;         // hit ratio (for split)
    ny::uint _lookup;      // lookup    (for pke)
    ny::uint _flen;        // traverse  (for pke)
    ny::uint _traverse;    // traverse  (for pke)
    ny::uint _lookup_sp;   // lookup    (for pke)
    ny::uint _traverse_sp; // traverse  (for pke)
#endif
    size_t _cost_fun (size_t m, size_t n);
    // bytewise coding for feature indexes (cf. Williams & Zobel, 1999)
    static void _vbEncode (ny::uint fi, ny::uchar* s, ny::uint &len);
    // various functions for sorting feature (index) vectors
    template <typename Iterator>
    static void _insertion_sort (const Iterator &first, const Iterator &last);
    template <typename value_type, typename SrcIterator, typename OutIterator>
    static void _radix (SrcIterator first, SrcIterator last, OutIterator dest, size_t shift);
    template <size_t SIZE, typename Iterator>
    static void _radix_sort (const Iterator &first, const Iterator &last);
    template <typename T, typename Iterator>
    static void _bucket_sort (const Iterator &first, const Iterator &last);
    // feature mapping
    void _convertCfstr2Cf (char* cs, ny::fv_t &fv, const bool flag);
    void _convertEx2Fv (char* cs, ny::fv_t &fv, const bool flag);
    void _convertFv2Fv (ny::fv_t &fnv) const;
    void _sortFv (ny::fv_t &fv) const;
    const ny::fv_it _splitFv (const ny::fv_t &fv) const;
    bool _renameFeatureID (std::vector <ny::fv_t> &fvv);
    // feature sequence trie builder
    virtual bool _abuse_double_array () const { return false; }
    std::string _setFStrieFN (const std::string &fn_base, const std::string &zfact);
    bool _setFStrie ();
    // classification functions
    void _addScore (ny::fl_t* score, int n);
    template <int D>
    void _pkeInnerLoop (ny::fl_t* score, ny::fv_it it, const ny::fv_it beg, const ny::fv_it end, const size_t pos);
    ny::fl_t* _pkeClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it &end);
    ny::fl_t* _pkeClassify (const ny::fv_t &fv, ny::fl_t* score)
    { return _pkeClassify (fv, score, fv.begin (), fv.end ()); }
    ny::fl_t* _fstClassify (const ny::fv_t &fv, ny::fl_t* score);
    virtual ny::fl_t* _baseClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it &end) = 0;
    int _lookupFeat (const ny::fv_it &it, size_t &pos);
    bool _setOpt (const char* opt_str);
  public:
    ClassifierBase () :
      _fv (), _opt ("pecco - please enjoy classification w/ conjunctive features\nCopyright (c) 2008-2010 Naoki Yoshinaga\n"), _mode (PKE), _verbosity (0), _d (1), _nl (0), _zmode (NFREQ_FCNT), _nf (0), _nf_cut (0), _nt (0), _nunit (0), _fn2fi (), _fi2fn (), _fncnt (), _pke (), _fst (), _feat_weight (0), _fst_weight (0), _f_r (0), _maf (0)
#ifdef USE_TIMER
      , _timer_pool (), _enc_t (_timer_pool.push ("enc")), _model_t (_timer_pool.push ("model")), _pke_t (_timer_pool.push ("pke", "classify")), _fst_t (_timer_pool.push ("fst", "classify"))
#endif
#ifdef USE_PROFILE
      , _all (0), _hit (0), _lookup (0), _flen (0), _traverse (0), _lookup_sp (0), _traverse_sp (0)
#endif
    {
      // options & default parametrs
      _opt.set ("-m",  "model", "model file");
      _opt.set ("-cf", "event", "event file to build fstrie");
#ifdef USE_FREQ_SORT
      _opt.set ("-lf", "train", "training file to rename features");
#endif
      _opt.set ("-tf", "test",  "test file; set '-' to read stdin");
      _opt.set ("-zt", "none",   "fstrie compression type; none / nff");
      _opt.set ("-z",  "0",     "fstrie compression factor (>0)");
      _opt.set ("-v",  "0",     "verbosity");
      
    }
    virtual ~ClassifierBase () {};
    // interface
    virtual bool load  () = 0;
    virtual bool batch () = 0; // pecco interface
    virtual ny::fl_t classify (ny::fv_t &fnv) = 0;
    void setOpt (const char* opt_key, const char* val) { _opt.set (opt_key, val); }
#ifdef USE_TIMER
    void printTimer () { _timer_pool.print (); }
#endif
#ifdef USE_PROFILE
    const void   printHitRatio () const;
#endif
  };
}
#endif /* CLASSIFY_H */
