// pecco -- please enjoy classification with conjunctive features
//  $Id: llm.h 472 2010-08-13 13:43:20Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef LLM_CLASSIFY_H
#define LLM_CLASSIFY_H

#include <sys/stat.h>
#include <cmath>
#include <cassert>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <set>
#include "typedef.h"
#include "opt_parse.h"
#ifdef USE_TIMER
#include "timer.h"
#endif
#include "classify.h"

// switch off
#undef ABUSE_DA

namespace pecco {
  class LLM : public ClassifierBase <MULTI> {
  private:
    // llm parameters
    ny::uint _tli;
    // feature weight
    std::vector <std::vector <ny::fl_t> > _alpha_tmp_ary; // alpha
    // mapping label to label index
    std::vector <char*>                           _li2l;  // label id -> label
    std::map <char*, ny::uint, ny::pless <char> > _l2li;  // label -> label id
    // temporary variables
    std::vector <ny::fv_t> _fc;
    // internal functions
    bool _setFtrie ();
    // classification
    ny::fl_t* _baseClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it &end)
      { return _pkeClassify (fv, score, it, end); }
  public:
    LLM (const char* opt_str) : ClassifierBase <MULTI> (), _tli (1), _alpha_tmp_ary (), _li2l (), _l2li () {
      _opt.set ("-c", "pke", "classifier: pke / fst");
      _setOpt (opt_str);
      if      (_opt["-c"] == "pke") _mode = PKE;
      else if (_opt["-c"] == "fst") _mode = FST;
      else ny::print_err (HERE "unknown classifier: %s\n", _opt["-c"].c_str ());
    }
    ~LLM () {
      for (size_t i = 0; i < _nl; ++i)  delete [] _li2l[i];
      if (_mode == PKE || _mode == FST) delete [] _feat_weight;
      if (_mode == FST)                 delete [] _fst_weight;
    }
    bool load  (); // set up model
    bool batch (); // pecco interface
    bool printParam () {
      std::cerr << "Degree: " << _d  << std::endl;
      std::cerr << "# of active features: " << _nf;
      if (_mode == PKE || _mode == FST) std::cerr << " (" << _nunit << ")";
      std::cerr << std::endl;
      return true;
    }
    ny::fl_t _calcProb (const ny::fl_t* const score, int tli_ = -1) const {
      ny::fl_t sum = 0;
      ny::fl_t prob_pos = 0;
      ny::uint tli = (tli_ == -1) ? _tli : static_cast <ny::uint> (tli_);
      for (size_t li = 0; li < _nl; ++li) {
        ny::fl_t prob = std::exp (score[li]);
        sum += prob;
        if (li == tli) prob_pos = prob;
      }
      return (prob_pos / sum);
    }
    // classification interface
    const ny::fl_t* classify (ny::fv_t &fnv, ny::fl_t* score) {
      std::fill (&score[0], &score[_nl], 0);
#ifdef USE_TIMER
      if (_verbosity > 0) _enc_t->startTimer ();
      _convertFv2Fv (fnv);
      if (_verbosity > 0) _enc_t->stopTimer ();
#else
      _convertFv2Fv (fnv);
#endif
      switch (_mode) {
      case PKE:
#ifdef USE_TIMER
        _pke_t->startTimer ();
        if (_d >= 2) _sortFv (fnv); _pkeClassify (fnv, &score[0]);
        _pke_t->stopTimer ();
#else
        if (_d >= 2) _sortFv (fnv); _pkeClassify (fnv, &score[0]);
#endif
        break;
      case FST:
#ifdef USE_TIMER
        _fst_t->startTimer ();
        _sortFv (fnv); _fstClassify (fnv, &score[0]);
        _fst_t->stopTimer ();
#else
        _sortFv (fnv); _fstClassify (fnv, &score[0]);
#endif
        break;
      default: ny::print_err (HERE "unknown classifier.\n");
      }
      return score;
    }
    ny::fl_t classify (ny::fv_t &fnv) {
      ny::fl_t * score = new ny::fl_t[_nl];
      ny::fl_t prob = _calcProb (classify (fnv, &score[0]));
      delete [] score;
      return prob;
    }
    bool binClassify (ny::fv_t &fnv, int tli_ = -1) {
      ny::uint tli = (tli_ == -1) ? _tli : static_cast <int> (tli_);
      ny::fl_t * score = new ny::fl_t[_nl];
      classify (fnv, &score[0]);
      delete [] score;
      return std::max_element (&score[0], &score[_nl]) == &score[tli];
    }
    const ny::fl_t* classify (char* cs, ny::fl_t* score) {
      // ny::fv_t fnv;
      _fv.clear ();
      for (cs = std::strtok (cs, " \t:"); cs != NULL; cs = std::strtok (NULL, " \t:")) { // " :"
        _fv.push_back (ny::strton <ny::uint> (cs));
        cs = std::strtok (NULL, " \t:"); // " :"
      }
      return classify (_fv, score);
    }
  };
}

#endif /* LLM_CLASSIFY_H */
