// pecco -- please enjoy classification with conjunctive features
//  $Id: polyk.h 503 2011-01-10 07:08:56Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef POLYK_CLASSIFY_H
#define POLYK_CLASSIFY_H

#include <sys/stat.h>
#include <cassert>
#include <cmath>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <set>
#include "typedef.h"
#ifdef USE_HASH
#include <tr1/unordered_map>
#endif
#include "opt_parse.h"
#ifdef USE_TIMER
#include "timer.h"
#endif
#include "classify.h"

namespace pecco {
  
  struct sv_t { // support vector
    ny::fv_t fv;
    ny::fl_t alpha;
    sv_t (const ny::fv_t &fv_, const ny::fl_t alpha_)
      : fv (fv_), alpha (alpha_) {};
    const ny::fv_it begin () const { return fv.begin (); }
    const ny::fv_it end ()   const { return fv.end (); }
  };
#ifdef ABUSE_DA
  class PolyK : public ClassifierBase <BINARY>
#else
  class PolyK : public ClassifierBase <MULTI>
#endif
  {
    // type alias
    typedef std::vector <ny::uint> ss_t;
    typedef std::vector <ny::uint>::const_iterator ss_it;
  private:
    // kernel parameters; (_s w * x + _r)^_d
    ny::fl_t _s;
    ny::fl_t _r;
    ny::fl_t _b;
    // support vector
    std::vector <ny::fl_t> _alph; // alpha used for PKI
    std::vector <sv_t>     _sv;   // support vector id -> <alpha, feature vector>
    std::vector <ss_t>     _f2ss; // feature to support vector ids <PKI>
    // size handler
    ny::uint _exnunit;            // # conjunctive features
    ny::uint _exnunit_cut;        // # conjunctive features (pruned)
    // kernel cache
    ny::fl_t*  _polyk;
    ny::fl_t*  _spolyk;
    // temporary variables
    ny::uint * _dot;
    bool*      _fbit;  // store feature on/off
    // polynomial kernel expansion
    ny::fl_t _sigma;
    ny::fl_t _sigma_neg;
    ny::fl_t _sigma_pos;
    ny::uint _processed;
    ny::fl_t _feat_cutoff_ratio;
    ny::fl_t _coeff[5]; // support up to d=4
    ny::fl_t _max_coeff;
    // timer
#ifdef USE_TIMER
#ifndef NDEBUG
    ny::Timer* _pkb_t;    // pkb
    ny::Timer* _pkbpp_t;  // pkb++ (bool* impl.)
#endif
    ny::Timer* _pki_t;    // pki
    ny::Timer* _split_t;  // split
#endif
    // internal functions
    bool _splitInit ();
    bool _precomputeKernel ();
    bool _setFtrie ();
    bool _setPKEcoeff ();
    void _pkePrefixSpan (ny::fv_t &fc, const std::vector <std::pair <ny::uint, int> > &proj, std::vector <FeatKey*> &pke_key);
    // classifier
#ifndef NDEBUG
    const ny::fl_t* _pkbClassify   (const ny::fv_t &fv, ny::fl_t* score) const;
    const ny::fl_t* _pkbppClassify (const ny::fv_t &fv, ny::fl_t* score);
#endif
    const ny::fl_t* _pkiClassify   (const ny::fv_t &fv, ny::fl_t* score) const;
    ny::fl_t* _splitClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it end);
    ny::fl_t* _splitClassify (const ny::fv_t &fv, ny::fl_t* score)
    { return _splitClassify (fv, score, fv.begin (), fv.end ()); }
#ifdef ABUSE_DA
    bool _abuse_double_array () const { return true; }
#endif
    ny::fl_t* _baseClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it it, const ny::fv_it &end)
    { return _splitClassify (fv, score, it, end); }
  public:
    PolyK (const char* opt_str) :
#ifdef ABUSE_DA
      ClassifierBase <BINARY> (),
#else
      ClassifierBase <MULTI> (),
#endif
      _s (0), _r (0), _b (0), _alph (), _sv (), _f2ss (), _exnunit (0), _exnunit_cut (0), _polyk (0), _spolyk (0), _dot (0), _fbit (0), _sigma (0), _sigma_neg (0), _sigma_pos (0), _processed (0), _feat_cutoff_ratio (0), _max_coeff (0)
#ifdef USE_TIMER
#ifndef NDEBUG
      , _pkb_t (_timer_pool.push ("pkb", "classify")), _pkbpp_t (_timer_pool.push ("pkbpp", "classify"))
#endif
      , _pki_t (_timer_pool.push ("pki", "classify")), _split_t (_timer_pool.push ("split", "classify")) 
#endif
      {
      _nl = 1; // bug
      _opt.set ("-c", "pke", "classifier: pki / pke / split / fst");
      _opt.set ("-r", "0", "splitSVM common-feature threshold");
      _opt.set ("-s", "0", "pke sigma");
      _setOpt (opt_str);
      // PKE sigma + splitSVM ratio
      _sigma = ny::strton <ny::fl_t> (_opt["-s"].c_str ());
      _feat_cutoff_ratio = ny::strton <ny::fl_t> (_opt["-r"].c_str ());
      std::string cfn = _opt["-cf"];
      _opt.set ("-cf", cfn);
      if      (_opt["-c"] == "pki" )  _mode = PKI;
      else if (_opt["-c"] == "pke")   _mode = PKE;
      else if (_opt["-c"] == "split") _mode = SPLIT;
      else if (_opt["-c"] == "fst")   _mode = FST;
#ifndef NDEBUG
      else if (_opt["-c"] == "pkb")   _mode = PKB;
      else if (_opt["-c"] == "pkbpp") _mode = PKBPP;
#endif
      else ny::print_err (HERE "unknown classifier: %s\n", _opt["-c"].c_str ());
    }
    ~PolyK () {
      if (_mode != PKE ) {
        delete [] _polyk;
        if (_mode == SPLIT || _mode == FST) delete [] _spolyk;
      }
      if (_mode == PKI)
        delete [] _dot;
#ifdef NDEBUG
      if (_mode != PKI)
#else
      if (_mode != PKB && _mode != PKI)
#endif
        delete [] _fbit;
#ifndef ABUSE_DA
      if (_mode == PKE || _mode == SPLIT || _mode == FST)
        delete [] _feat_weight;
      if (_mode == FST)
        delete [] _fst_weight;
#endif
    }
    const std::string& getSigma () { return _opt["-s"]; };
    bool load  (); // set up model
    bool batch (); // pecco interface
    bool printParam () {
      std::cerr << "Kernel: " << "(" << _s << "*wx+" << _r << ")^"
                << _d << std::endl;
      std::cerr << "# support vectors: " << _nunit << std::endl;
      std::cerr << "# of active features: " << _nf_cut;
      if (_mode == PKE || _mode == SPLIT || _mode == FST)
        std::cerr << " (" << _exnunit_cut << ")";
      std::cerr << std::endl;
      if (_verbosity > 1)
        std::cerr << "  # common feature: " << _f_r-1 << std::endl;
      return true;
    }
    // classification interface
    ny::fl_t classify (ny::fv_t &fnv) {
#ifdef USE_TIMER
      if (_verbosity > 0) _enc_t->startTimer ();
      _convertFv2Fv (fnv);
      if (_verbosity > 0) _enc_t->stopTimer ();
#else
      _convertFv2Fv (fnv);
#endif
      ny::fl_t w = -_b;
      switch (_mode) {
      case PKI:
#ifdef USE_TIMER
        _pki_t->startTimer ();
        _pkiClassify (fnv, &w);
        _pki_t->stopTimer ();
#else
	_pkiClassify (fnv, &w);
#endif
      break;
      case PKE:
#ifdef USE_TIMER
       _pke_t->startTimer ();
       if (_d >= 2) _sortFv (fnv); _pkeClassify (fnv, &w);
       _pke_t->stopTimer ();
#else
       if (_d >= 2) _sortFv (fnv); _pkeClassify (fnv, &w);
#endif
        break;
      case SPLIT:
#ifdef USE_TIMER
        _split_t->startTimer ();
        _sortFv (fnv); _splitClassify (fnv, &w);
        _split_t->stopTimer ();
#else
	_sortFv (fnv); _splitClassify (fnv, &w);
#endif
        break;
      case FST:
#ifdef USE_TIMER
        _fst_t->startTimer ();
        _sortFv (fnv); _fstClassify (fnv, &w);
        _fst_t->stopTimer ();
#else
	_sortFv (fnv); _fstClassify (fnv, &w);
#endif
        break;
#ifndef NDEBUG
      case PKB:
#ifdef USE_TIMER
        _pkb_t->startTimer ();
        _sortFv (fnv); _pkbClassify (fnv, &w);
        _pkb_t->stopTimer ();
#else
	_sortFv (fnv); _pkbClassify (fnv, &w);
#endif
        break;
      case PKBPP:
#ifdef USE_TIMER
        _pkbpp_t->startTimer ();
        _pkbppClassify (fnv, &w);
        _pkbpp_t->stopTimer ();
#else
	_pkbppClassify (fnv, &w);
#endif
        break;
#endif
      default: ny::print_err (HERE "unknown classifier.\n");
      }
      return w;
    }
    bool binClassify (ny::fv_t &fnv) { return (classify (fnv) > 0); }
    ny::fl_t classify (char* cs) {
      // ny::fv_t fnv;
      _fv.clear ();
      for (cs = std::strtok (cs, " \t:"); cs != NULL; cs = std::strtok (NULL, " \t:")) { // " :"
        _fv.push_back (ny::strton <ny::uint> (cs));
        cs = std::strtok (NULL, " \t:");  // " :"
      }
      return classify (_fv);
    }
    // helper function
    static ny::fl_t sigmoid (ny::fl_t x)
    { return static_cast <ny::fl_t> (1.0 / (1.0 + std::exp (-x))); }
  };
}
#endif /* POLYK_CLASSIFY_H */
