// pecco -- please enjoy classification with conjunctive features
//  $Id: typedef.h 472 2010-08-13 13:43:20Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef TYPEDEF_H
#define TYPEDEF_H

#include <sys/stat.h>
#include <stdint.h>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cstdarg>
#include <iostream>
#include <string>
#include <vector>
#include "pecco_conf.h"
#ifdef USE_HASH
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#else
#include <map>
#include <set>
#endif
#ifdef USE_DARTS
#include <darts.h>
#endif
#ifdef USE_DARTS_CLONE
#include <darts-clone.h>
#endif
#ifdef USE_DDA
#include "dda.h"
#endif

#define _STR_(x)   #x
#define _STR(x)    _STR_(x)
#define HERE       __FILE__ " [" _STR(__LINE__) "]: "

namespace ny {
#ifdef USE_DARTS
  static const std::string DA_SUFFIX = "da";
#endif
#ifdef USE_DARTS_CLONE
  static const std::string DA_SUFFIX = "dan";
#endif
#ifdef USE_DDA
  static const std::string DA_SUFFIX = "dda";
#endif
  // type alias
  // ABUSE_DA requires 32-bit fl_t
#if defined (USE_FLOAT) || defined (ABUSE_DA)
  typedef float  fl_t;
#else
  typedef double fl_t;
#endif
  typedef unsigned int         uint;
  typedef unsigned char        uchar;
#ifdef USE_DDA
  typedef pda::da <int, -1, -2, 32>  DA;
#else
  typedef Darts::DoubleArray   DA;
#endif
  typedef std::vector <uint>   fv_t;
  typedef fv_t::const_iterator fv_it;
  // template typedef
#ifdef USE_HASH
  template <typename _Key, typename _Tp>
  struct map { typedef std::tr1::unordered_map <_Key, _Tp> type; };
  template <typename _Key>
  struct set { typedef std::tr1::unordered_set <_Key> type; };
#else
  template <typename _Key, typename _Tp>
  struct map { typedef std::map <_Key, _Tp> type; };
  template <typename _Key>
  struct set { typedef std::set <_Key> type; };
#endif
  template <typename T>
  struct counter { typedef std::vector <std::pair <ny::uint, T> > type; };
  // uncopyable mix-in
  class Uncopyable {
  protected:
    Uncopyable() {}
    ~Uncopyable() {}
  private:
    Uncopyable (const Uncopyable&);
    Uncopyable& operator= (const Uncopyable&);
  };
  // less for pointers
  template <typename T>
  struct pless {
    bool operator () (const T* a, const T* b) const {
      while (1) {
        if      (*a == '\0') return (*b != '\0');
        else if (*b == '\0') return false;
        else if (static_cast <uint> (*a) > static_cast <uint> (*b))
          return false;
        else if (static_cast <uint> (*a) < static_cast <uint> (*b))
          return true;
        ++a; ++b;
      }
    }
  };
  // double array keys; key must be terminated with '\0'
  template <typename T, typename U>
  class DAKeyBase : private Uncopyable {
  public:
    T*     key;
    size_t len;
    U*     cont;
    DAKeyBase (T* k, U* c, size_t l = 0, size_t nl = 1)
      : key (NULL), len (l ? l : length (k)), cont (NULL) {
      key = new T[len + 1];
      key[len] = '\0'; // required for i386
      std::memcpy (key, k, len * sizeof (T));
      cont = new U[nl + 1];
      if (c) std::memcpy (cont, c, nl * sizeof (U));
    };
    size_t length (const T* k) const {
      const T* p = k;
      while (*p != static_cast<T> (0)) ++p;
      return p - k;
    };
    bool operator<  (const DAKeyBase &a) const
    { return pless <T> () (key, a.key); }
    ~DAKeyBase () {
      delete [] key;
      delete [] cont;
    }
  };
  //
  template <typename T, typename U>
  struct DAKeypLess {
    typedef DAKeyBase <T, U> Key;
    bool operator () (const Key* a, const Key* b) const
    { return pless <T> () (a->key, b->key); }
  };
  // getline wrapper
  static inline bool getLine (FILE* &fp, char* &line, size_t &read) {
#ifdef __APPLE__
    if ((line = fgetln (fp, &read)) == NULL) return false;
#else
    static ssize_t read_ = 0; static size_t size = 0; // static for inline
    if ((read_ = getline (&line, &size, fp)) == -1) return false;
    read = read_;
#endif
    *(line + read - 1) = '\0';
    return true;
  }
  static inline void print_err (char const *format, ...) {
    std::va_list argptr;
    va_start (argptr, format);
    vfprintf (stderr, format, argptr);
    va_end   (argptr);
    std::exit (1);
  }
  // replace the first char c in s with '\0' and return a pointer to the next
  static inline char* skipChar (char* s, const char c) {
    char *p = s;
    while (*p != '\0' && *p != c) ++p;
    if (*p == '\0') print_err ("string doesn't contain %c\n", c);
    *p = '\0';
    return ++p;
  }
  static inline char* skipChar (char* s, const char c, const char* s_end) {
    char *p = s;
    while (p != s_end && *p != c) ++p;
    *p = '\0';
    return ++p;
  }
  // str to numerical wrapper
  template <typename T> T strton (const char* s, char** error);
  template <typename T> T strton (const char* s);
}
#endif /* TYPEDEF_H */
